#!/usr/bin/env python
# License: GPLv3 Copyright: 2014, Kovid Goyal <kovid at kovidgoyal.net>


class NullSmarts:
    override_tab_stop_width = None

    def __init__(self, editor):
        pass

    def get_extra_selections(self, editor):
        return ()

    def get_smart_selection(self, editor, update=True):
        return editor.selected_text

    def verify_for_spellcheck(self, cursor, highlighter):
        return False

    def cursor_position_with_sourceline(self, cursor, for_position_sync=True, use_matched_tag=True):
        return None, None

    def goto_sourceline(self, editor, sourceline, tags, attribute=None):
        return False

    def get_inner_HTML(self, editor):
        return None

    def handle_key_press(self, ev, editor):
        return False

    def get_completion_data(self, editor, ev=None):
        return None

    def rename_block_tag(self, editor, new_name):
        pass

    def set_text_alignment(self, editor, value):
        pass

    def surround_with_custom_tag(self, editor, opent, close):
        pass

    def insert_hyperlink(self, editor, target, text, template=None):
        pass

    def insert_tag(self, editor, name):
        pass

    def remove_tag(self, editor):
        pass

    def split_tag(self, editor):
        pass



#!/usr/bin/env python
# License: GPLv3 Copyright: 2009, Kovid Goyal <kovid@kovidgoyal.net>


class Clean:
    """Clean up guide, leaving only known values"""

    def __call__(self, oeb, opts):
        self.oeb, self.log, self.opts = oeb, oeb.log, opts

        if 'cover' not in self.oeb.guide:
            covers = []
            for x in (
                'other.ms-coverimage-standard',
                'coverimagestandard',
                'other.ms-titleimage-standard',
                'other.ms-titleimage',
                'other.ms-coverimage',
                'other.ms-thumbimage-standard',
                'other.ms-thumbimage',
                'thumbimagestandard',
            ):
                if x in self.oeb.guide:
                    href = self.oeb.guide[x].href
                    try:
                        item = self.oeb.manifest.hrefs[href]
                    except KeyError:
                        continue
                    else:
                        covers.append((self.oeb.guide[x], len(item.data)))

            covers.sort(key=lambda x: x[1], reverse=True)
            if covers:
                ref = covers[0][0]
                if len(covers) > 1:
                    self.log(f'Choosing {ref.type}:{ref.href} as the cover')
                ref.type = 'cover'
                self.oeb.guide.refs['cover'] = ref

        if 'start' in self.oeb.guide and 'text' not in self.oeb.guide:
            # Prefer text to start as per the OPF 2.0 spec
            x = self.oeb.guide['start']
            self.oeb.guide.add('text', x.title, x.href)
            self.oeb.guide.remove('start')

        for x in list(self.oeb.guide):
            if x.lower() not in {
                'cover',
                'titlepage',
                'masthead',
                'toc',
                'title-page',
                'copyright-page',
                'text',
                'index',
                'glossary',
                'acknowledgements',
                'bibliography',
                'colophon',
                'copyright-page',
                'dedication',
                'epigraph',
                'foreword',
                'loi',
                'lot',
                'notes',
                'preface',
            }:
                item = self.oeb.guide[x]
                if item.title and item.title.lower() == 'start':
                    continue
                self.oeb.guide.remove(x)



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid@kovidgoyal.net>

import os

from calibre import replace_entities
from calibre.ebooks.metadata.toc import TOC
from calibre.ebooks.mobi.reader.headers import NULL_INDEX
from calibre.ebooks.mobi.reader.index import read_index

tag_fieldname_map = {
    1: ['pos', 0],
    2: ['len', 0],
    3: ['noffs', 0],
    4: ['hlvl', 0],
    5: ['koffs', 0],
    6: ['pos_fid', 0],
    21: ['parent', 0],
    22: ['child1', 0],
    23: ['childn', 0],
    69: ['image_index', 0],
    70: ['desc_offset', 0],  # 'Description offset in cncx'
    71: ['author_offset', 0],  # 'Author offset in cncx'
    72: ['image_caption_offset', 0],  # 'Image caption offset in cncx',
    73: ['image_attr_offset', 0],  # 'Image attribution offset in cncx',
}

default_entry = {
    'pos': -1,
    'len': 0,
    'noffs': -1,
    'text': 'Unknown Text',
    'hlvl': -1,
    'kind': 'Unknown Class',
    'pos_fid': None,
    'parent': -1,
    'child1': -1,
    'childn': -1,
    'description': None,
    'author': None,
    'image_caption': None,
    'image_attribution': None,
}


def read_ncx(sections, index, codec):
    index_entries = []

    if index != NULL_INDEX:
        table, cncx = read_index(sections, index, codec)

        for num, x in enumerate(table.items()):
            text, tag_map = x
            entry: dict = default_entry.copy()
            entry['name'] = text
            entry['num'] = num

            for tag, (fieldname, i) in tag_fieldname_map.items():
                if tag in tag_map:
                    fieldvalue = tag_map[tag][i]
                    if tag == 6:
                        # Appears to be an idx into the KF8 elems table with an
                        # offset
                        fieldvalue = tuple(tag_map[tag])
                    entry[fieldname] = fieldvalue
                    for which, name in {
                        3: 'text',
                        5: 'kind',
                        70: 'description',
                        71: 'author',
                        72: 'image_caption',
                        73: 'image_attribution',
                    }.items():
                        if tag == which:
                            entry[name] = cncx.get(fieldvalue, default_entry[name])
            index_entries.append(entry)

    return index_entries


def build_toc(index_entries):
    ans = TOC(base_path=os.getcwd())
    levels = {x['hlvl'] for x in index_entries}
    num_map = {-1: ans}
    level_map = {l: [x for x in index_entries if x['hlvl'] == l] for l in levels}
    for lvl in sorted(levels):
        for item in level_map[lvl]:
            parent = num_map[item['parent']]
            child = parent.add_item(item['href'], item['idtag'], replace_entities(item['text'], encoding=None))
            num_map[item['num']] = child

    # Set play orders in depth first order
    for i, item in enumerate(ans.flat()):
        item.play_order = i

    return ans



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

import re
import unicodedata

import html5_parser
from lxml.etree import Element as LxmlElement

from calibre.ebooks.chardet import strip_encoding_declarations, xml_to_unicode
from calibre.utils.cleantext import clean_xml_chars
from calibre.utils.xml_parse import safe_xml_fromstring
from calibre_extensions.fast_html_entities import replace_all_entities

XHTML_NS = 'http://www.w3.org/1999/xhtml'


def decode_xml(data: bytes, normalize_to_nfc: bool = True) -> tuple[str, str]:
    used_encoding = ''

    def fix_data(d):
        return d.replace('\r\n', '\n').replace('\r', '\n'), used_encoding

    if isinstance(data, str):
        return fix_data(data)
    bom_enc = None
    if data[:4] in {b'\0\0\xfe\xff', b'\xff\xfe\0\0'}:
        bom_enc = {b'\0\0\xfe\xff': 'utf-32-be', b'\xff\xfe\0\0': 'utf-32-le'}[data[:4]]
        data = data[4:]
    elif data[:2] in {b'\xff\xfe', b'\xfe\xff'}:
        bom_enc = {b'\xff\xfe': 'utf-16-le', b'\xfe\xff': 'utf-16-be'}[data[:2]]
        data = data[2:]
    elif data[:3] == b'\xef\xbb\xbf':
        bom_enc = 'utf-8'
        data = data[3:]
    if bom_enc is not None:
        try:
            used_encoding = bom_enc
            return fix_data(data.decode(bom_enc))
        except UnicodeDecodeError:
            pass
    try:
        used_encoding = 'utf-8'
        return fix_data(data.decode('utf-8'))
    except UnicodeDecodeError:
        pass
    str_data, used_encoding = xml_to_unicode(data)
    if normalize_to_nfc:
        str_data = unicodedata.normalize('NFC', str_data)
    return fix_data(str_data)


def parse_html5(
    raw,
    decoder=None,
    log=None,
    discard_namespaces=False,
    line_numbers=True,
    linenumber_attribute=None,
    replace_entities=True,
    fix_newlines=True,
):
    if isinstance(raw, bytes):
        raw = xml_to_unicode(raw)[0] if decoder is None else decoder(raw)
    assert isinstance(raw, str)
    if replace_entities:
        raw = replace_all_entities(raw, True)
        assert isinstance(raw, str)
    if fix_newlines:
        raw = raw.replace('\r\n', '\n').replace('\r', '\n')
    raw = clean_xml_chars(raw)
    root = html5_parser.parse(
        raw,
        maybe_xhtml=not discard_namespaces,
        line_number_attr=linenumber_attribute,
        keep_doctype=False,
        sanitize_names=True,
    )
    if (discard_namespaces and root.tag != 'html') or (not discard_namespaces and (root.tag != '{{{}}}{}'.format(XHTML_NS, 'html') or root.prefix)):
        raise ValueError(f'Failed to parse correctly, root has tag: {root.tag} and prefix: {root.prefix}')
    return root


def handle_private_entities(data):
    # Process private entities
    pre = ''
    idx = data.find('<html')
    if idx == -1:
        idx = data.find('<HTML')
    if idx > -1:
        pre = data[:idx]
        num_of_nl_in_pre = pre.count('\n')
        if '<!DOCTYPE' in pre:  # Handle user defined entities
            user_entities = {}
            for match in re.finditer(r'<!ENTITY\s+(\S+)\s+([^>]+)', pre):
                val = match.group(2)
                if val.startswith('"') and val.endswith('"'):
                    val = val[1:-1]
                user_entities[match.group(1)] = val
            if user_entities:
                data = ('\n' * num_of_nl_in_pre) + data[idx:]
                pat = re.compile(r'&({});'.format('|'.join(user_entities.keys())))
                data = pat.sub(lambda m: user_entities[m.group(1)], data)
    return data


def parse(
    raw,
    decoder=None,
    log=None,
    line_numbers=True,
    linenumber_attribute=None,
    replace_entities=True,
    force_html5_parse=False,
):
    if isinstance(raw, bytes):
        raw = xml_to_unicode(raw)[0] if decoder is None else decoder(raw)
    assert isinstance(raw, str)
    raw = handle_private_entities(raw)
    assert isinstance(raw, str)
    if replace_entities:
        raw = replace_all_entities(raw, True)
        assert isinstance(raw, str)
    raw = raw.replace('\r\n', '\n').replace('\r', '\n')

    # Remove any preamble before the opening html tag as it can cause problems,
    # especially doctypes, preserve the original linenumbers by inserting
    # newlines at the start
    pre = raw[:2048]
    for match in re.finditer(r'<\s*html', pre, flags=re.I):
        newlines = raw.count('\n', 0, match.start())
        raw = ('\n' * newlines) + raw[match.start() :]
        break

    raw = strip_encoding_declarations(raw, limit=10 * 1024, preserve_newlines=True)
    if force_html5_parse:
        return parse_html5(
            raw,
            log=log,
            line_numbers=line_numbers,
            linenumber_attribute=linenumber_attribute,
            replace_entities=False,
            fix_newlines=False,
        )
    try:
        ans = safe_xml_fromstring(raw, recover=False)
        if ans.tag != f'{{{XHTML_NS}}}html':
            raise ValueError('Root tag is not <html> in the XHTML namespace')
        if linenumber_attribute:
            for elem in ans.iter(LxmlElement):
                if elem.sourceline is not None:
                    elem.set(linenumber_attribute, str(elem.sourceline))
        return ans
    except Exception:
        if log is not None:
            log.exception('Failed to parse as XML, parsing as tag soup')
        return parse_html5(
            raw,
            log=log,
            line_numbers=line_numbers,
            linenumber_attribute=linenumber_attribute,
            replace_entities=False,
            fix_newlines=False,
        )


if __name__ == '__main__':
    from lxml import etree

    root = parse_html5(
        '\n<html><head><title>a\n</title><p b=1 c=2 a=0>&nbsp;\n<b>b<svg ass="wipe" viewbox="0">',
        discard_namespaces=False,
    )
    print(etree.tostring(root, encoding='utf-8'))
    print()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2025, Kovid Goyal <kovid at kovidgoyal.net>

from calibre.customize import AIProviderPlugin
from calibre.utils.localization import _


class OpenAI(AIProviderPlugin):
    name = 'OpenAI'
    version = (1, 0, 0)
    description = _('AI services from OpenAI')
    author = 'Kovid Goyal'
    builtin_live_module_name = 'calibre.ai.openai.backend'

    @property
    def capabilities(self):
        from calibre.ai import AICapabilities

        return (
            AICapabilities.text_to_text | AICapabilities.embedding | AICapabilities.text_and_image_to_image | AICapabilities.text_to_image | AICapabilities.tts
        )



#!/usr/bin/env python
# License: GPLv3 Copyright: 2010, Kovid Goyal <kovid@kovidgoyal.net>

import re
from functools import partial

from qt.core import QDropEvent, QListWidget, QListWidgetItem, Qt

from calibre.constants import iswindows
from calibre.customize.ui import all_input_formats, available_output_formats
from calibre.ebooks import BOOK_EXTENSIONS
from calibre.ebooks.oeb.iterator import is_supported
from calibre.gui2 import config, dynamic, gprefs, info_dialog
from calibre.gui2.actions.choose_library import get_change_library_action_plugin
from calibre.gui2.preferences import ConfigWidgetBase, Setting, get_move_count, test_widget
from calibre.gui2.preferences.behavior_ui import Ui_Form
from calibre.utils.config import prefs
from calibre.utils.icu import sort_key
from calibre.utils.localization import _


class OutputFormatSetting(Setting):
    CHOICES_SEARCH_FLAGS = Qt.MatchFlag.MatchFixedString


class ConfigWidget(ConfigWidgetBase, Ui_Form):
    def genesis(self, gui):
        self.gui = gui
        db = gui.library_view.model().db

        r = self.register
        choices = (
            [(_('Low'), 'low'), (_('Normal'), 'normal'), (_('High'), 'high')]
            if iswindows
            else [(_('Normal'), 'normal'), (_('Low'), 'low'), (_('Very low'), 'high')]
        )
        r('worker_process_priority', prefs, choices=choices)

        r('network_timeout', prefs)

        r('new_version_notification', config)
        r('upload_news_to_device', config)
        r('delete_news_from_library_on_upload', config)

        output_formats = sorted(available_output_formats())
        output_formats.remove('oeb')
        choices = [(x.upper(), x) for x in output_formats]
        r('output_format', prefs, choices=choices, setting=OutputFormatSetting)

        restrictions = sorted(db.prefs['virtual_libraries'], key=sort_key)
        choices = [('', '')] + [(x, x) for x in restrictions]
        # check that the virtual library still exists
        vls = db.prefs['virtual_lib_on_startup']
        if vls and vls not in restrictions:
            db.prefs['virtual_lib_on_startup'] = ''
        r('virtual_lib_on_startup', db.prefs, choices=choices)
        self.reset_confirmation_button.clicked.connect(self.reset_confirmation_dialogs)

        self.input_up_button.clicked.connect(partial(self.up_input, use_kbd_modifiers=True))
        self.input_down_button.clicked.connect(partial(self.down_input, use_kbd_modifiers=True))
        self.opt_input_order.set_movement_functions(partial(self.up_input, use_kbd_modifiers=False), partial(self.down_input, use_kbd_modifiers=False))
        self.opt_input_order.handle_drop_event = self.input_order_drop_event
        for signal in ('Activated', 'Changed', 'DoubleClicked', 'Clicked'):
            signal = getattr(self.opt_internally_viewed_formats, 'item' + signal)
            signal.connect(self.internally_viewed_formats_changed)

        r('bools_are_tristate', db.prefs, restart_required=True)
        r('numeric_collation', prefs, restart_required=True)

    def input_order_drop_event(self, ev: QDropEvent | None) -> None:
        QListWidget.dropEvent(self.opt_input_order, ev)
        if ev is not None and ev.isAccepted():
            self.changed_signal.emit()

    def initialize(self):
        ConfigWidgetBase.initialize(self)
        self.init_input_order()
        self.init_internally_viewed_formats()

    def restore_defaults(self, *args):
        ConfigWidgetBase.restore_defaults(self)
        self.init_input_order(defaults=True)
        self.init_internally_viewed_formats(defaults=True)
        self.changed_signal.emit()

    def commit(self, *args):
        input_map = prefs['input_format_order']
        input_cols = []
        for i in range(self.opt_input_order.count()):
            _item = self.opt_input_order.item(i)
            assert _item is not None
            input_cols.append(str(_item.data(Qt.ItemDataRole.UserRole) or ''))
        if input_map != input_cols:
            prefs['input_format_order'] = input_cols
        fmts = self.current_internally_viewed_formats
        old = config['internally_viewed_formats']
        if fmts != old:
            config['internally_viewed_formats'] = fmts
        ret = ConfigWidgetBase.commit(self)
        # Signal a possible change of the VL at startup opt
        get_change_library_action_plugin().rebuild_change_library_menus.emit()
        return ret

    # Internally viewed formats {{{
    def internally_viewed_formats_changed(self, *args):
        fmts = self.current_internally_viewed_formats
        old = config['internally_viewed_formats']
        if fmts != old:
            self.changed_signal.emit()

    def init_internally_viewed_formats(self, defaults=False):
        if defaults:
            fmts = config.defaults['internally_viewed_formats']
        else:
            fmts = config['internally_viewed_formats']
        viewer = self.opt_internally_viewed_formats
        viewer.blockSignals(True)
        exts = set()
        for ext in BOOK_EXTENSIONS:
            ext = ext.lower()
            ext = re.sub(r'(x{0,1})htm(l{0,1})', 'html', ext)
            if ext == 'lrf' or is_supported('book.' + ext):
                exts.add(ext)
        viewer.clear()
        for ext in sorted(exts):
            viewer.addItem(ext.upper())
            item = viewer.item(viewer.count() - 1)
            assert item is not None
            item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable)
            item.setCheckState(Qt.CheckState.Checked if ext.upper() in fmts else Qt.CheckState.Unchecked)
        viewer.blockSignals(False)

    @property
    def current_internally_viewed_formats(self):
        fmts = []
        viewer = self.opt_internally_viewed_formats
        for i in range(viewer.count()):
            _item = viewer.item(i)
            assert _item is not None
            if _item.checkState() == Qt.CheckState.Checked:
                fmts.append(str(_item.text()))
        return fmts

    # }}}

    # Input format order {{{
    def init_input_order(self, defaults=False):
        if defaults:
            input_map = prefs.defaults['input_format_order']
        else:
            input_map = prefs['input_format_order']
        all_formats = set()
        self.opt_input_order.clear()
        for fmt in all_input_formats().union({'ZIP', 'RAR'}):
            all_formats.add(fmt.upper())
        for format in input_map + list(all_formats.difference(input_map)):
            item = QListWidgetItem(format, self.opt_input_order)
            item.setData(Qt.ItemDataRole.UserRole, (format))
            item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsDragEnabled)

    def up_input(self, use_kbd_modifiers, *args):
        count = get_move_count(self.opt_input_order.count()) if use_kbd_modifiers else 1
        for _x in range(count):
            idx = self.opt_input_order.currentRow()
            if idx > 0:
                self.opt_input_order.insertItem(idx - 1, self.opt_input_order.takeItem(idx))
                self.opt_input_order.setCurrentRow(idx - 1)
                self.changed_signal.emit()

    def down_input(self, use_kbd_modifiers, *args):
        count = get_move_count(self.opt_input_order.count()) if use_kbd_modifiers else 1
        for _x in range(count):
            idx = self.opt_input_order.currentRow()
            if idx < self.opt_input_order.count() - 1:
                self.opt_input_order.insertItem(idx + 1, self.opt_input_order.takeItem(idx))
                self.opt_input_order.setCurrentRow(idx + 1)
                self.changed_signal.emit()

    # }}}

    def reset_confirmation_dialogs(self, *args):
        from calibre.gui2.tweak_book import tprefs
        from calibre.gui2.viewer.config import vprefs

        for confloc in (dynamic, vprefs, tprefs):
            for key in confloc.keys():
                if key.endswith('_again') and confloc[key] is False:
                    confloc[key] = True
        gprefs['questions_to_auto_skip'] = []
        info_dialog(self, _('Done'), _('Confirmation dialogs have all been reset'), show=True)


if __name__ == '__main__':
    from calibre.gui2 import Application

    app = Application([])
    test_widget('Interface', 'Behavior')



#!/usr/bin/env python
# License: GPLv3 Copyright: 2013, Kovid Goyal <kovid at kovidgoyal.net>

import numbers
from functools import wraps

from css_parser import CSSParser
from css_parser import profile as cssprofiles
from css_parser.css import PropertyValue
from tinycss.fonts3 import parse_font, serialize_font_family

from calibre.ebooks.oeb.base import css_text

DEFAULTS = {
    'azimuth': 'center',
    'background-attachment': 'scroll',  # {{{
    'background-color': 'transparent',
    'background-image': 'none',
    'background-position': '0% 0%',
    'background-repeat': 'repeat',
    'border-bottom-color': 'currentColor',
    'border-bottom-style': 'none',
    'border-bottom-width': 'medium',
    'border-collapse': 'separate',
    'border-left-color': 'currentColor',
    'border-left-style': 'none',
    'border-left-width': 'medium',
    'border-right-color': 'currentColor',
    'border-right-style': 'none',
    'border-right-width': 'medium',
    'border-spacing': 0,
    'border-top-color': 'currentColor',
    'border-top-style': 'none',
    'border-top-width': 'medium',
    'bottom': 'auto',
    'caption-side': 'top',
    'clear': 'none',
    'clip': 'auto',
    'color': 'black',
    'content': 'normal',
    'counter-increment': 'none',
    'counter-reset': 'none',
    'cue-after': 'none',
    'cue-before': 'none',
    'cursor': 'auto',
    'direction': 'ltr',
    'display': 'inline',
    'elevation': 'level',
    'empty-cells': 'show',
    'float': 'none',
    'font-family': 'serif',
    'font-size': 'medium',
    'font-stretch': 'normal',
    'font-style': 'normal',
    'font-variant': 'normal',
    'font-weight': 'normal',
    'height': 'auto',
    'left': 'auto',
    'letter-spacing': 'normal',
    'line-height': 'normal',
    'list-style-image': 'none',
    'list-style-position': 'outside',
    'list-style-type': 'disc',
    'margin-bottom': 0,
    'margin-left': 0,
    'margin-right': 0,
    'margin-top': 0,
    'max-height': 'none',
    'max-width': 'none',
    'min-height': 0,
    'min-width': 0,
    'orphans': '2',
    'outline-color': 'invert',
    'outline-style': 'none',
    'outline-width': 'medium',
    'overflow': 'visible',
    'padding-bottom': 0,
    'padding-left': 0,
    'padding-right': 0,
    'padding-top': 0,
    'page-break-after': 'auto',
    'page-break-before': 'auto',
    'page-break-inside': 'auto',
    'pause-after': 0,
    'pause-before': 0,
    'pitch': 'medium',
    'pitch-range': '50',
    'play-during': 'auto',
    'position': 'static',
    'quotes': "'“' '”' '‘' '’'",
    'richness': '50',
    'right': 'auto',
    'speak': 'normal',
    'speak-header': 'once',
    'speak-numeral': 'continuous',
    'speak-punctuation': 'none',
    'speech-rate': 'medium',
    'stress': '50',
    'table-layout': 'auto',
    'text-align': 'auto',
    'text-decoration': 'none',
    'text-indent': 0,
    'text-shadow': 'none',
    'text-transform': 'none',
    'top': 'auto',
    'unicode-bidi': 'normal',
    'vertical-align': 'baseline',
    'visibility': 'visible',
    'voice-family': 'default',
    'volume': 'medium',
    'white-space': 'normal',
    'widows': '2',
    'width': 'auto',
    'word-spacing': 'normal',
    'z-index': 'auto',
}
# }}}

EDGES = ('top', 'right', 'bottom', 'left')
BORDER_PROPS = ('color', 'style', 'width')


def normalize_edge(name, cssvalue):
    style = {}
    if isinstance(cssvalue, PropertyValue):
        primitives = [css_text(v) for v in cssvalue]
    else:
        primitives = [css_text(cssvalue)]
    if len(primitives) == 1:
        (value,) = primitives
        values = (value, value, value, value)
    elif len(primitives) == 2:
        vert, horiz = primitives
        values = (vert, horiz, vert, horiz)
    elif len(primitives) == 3:
        top, horiz, bottom = primitives
        values = (top, horiz, bottom, horiz)
    else:
        values = primitives[:4]
    if '-' in name:
        l, _, r = name.partition('-')
        for edge, value in zip(EDGES, values):
            style[f'{l}-{edge}-{r}'] = value
    else:
        for edge, value in zip(EDGES, values):
            style[f'{name}-{edge}'] = value
    return style


def simple_normalizer(prefix, names, check_inherit=True):
    composition = tuple(f'{prefix}-{n}' for n in names)

    @wraps(normalize_simple_composition)
    def wrapper(name, cssvalue):
        return normalize_simple_composition(name, cssvalue, composition, check_inherit=check_inherit)

    return wrapper


def normalize_simple_composition(name, cssvalue, composition, check_inherit=True):
    if check_inherit and css_text(cssvalue) == 'inherit':
        style = {k: 'inherit' for k in composition}
    else:
        style = {k: DEFAULTS[k] for k in composition}
        try:
            primitives = [css_text(v) for v in cssvalue]
        except TypeError:
            primitives = [css_text(cssvalue)]
        while primitives:
            value = primitives.pop()
            for key in composition:
                if cssprofiles.validate(key, value):
                    style[key] = value
                    break
    return style


font_composition = ('font-style', 'font-variant', 'font-weight', 'font-size', 'line-height', 'font-family')


def normalize_font(cssvalue, font_family_as_list=False):
    # See https://developer.mozilla.org/en-US/docs/Web/CSS/font
    composition = font_composition
    val = css_text(cssvalue)
    ans: dict
    if val == 'inherit':
        ans = {k: 'inherit' for k in composition}
    elif val in {'caption', 'icon', 'menu', 'message-box', 'small-caption', 'status-bar'}:
        ans = {k: DEFAULTS[k] for k in composition}
    else:
        ans = {k: DEFAULTS[k] for k in composition}
        ans.update(parse_font(val))
    if font_family_as_list:
        if isinstance(ans['font-family'], (str, bytes)):
            ans['font-family'] = [x.strip() for x in ans['font-family'].split(',')]
    elif not isinstance(ans['font-family'], (str, bytes)):
        ans['font-family'] = serialize_font_family(ans['font-family'])
    return ans


def normalize_border(name, cssvalue):
    style = normalizers['border-' + EDGES[0]]('border-' + EDGES[0], cssvalue)
    vals = style.copy()
    for edge in EDGES[1:]:
        style.update({k.replace(EDGES[0], edge): v for k, v in vals.items()})
    return style


normalizers = {
    'list-style': simple_normalizer('list-style', ('type', 'position', 'image')),
    'font': lambda prop, v: normalize_font(v),
    'border': normalize_border,
}

for x in ('margin', 'padding', 'border-style', 'border-width', 'border-color'):
    normalizers[x] = normalize_edge

for x in EDGES:
    name = 'border-' + x
    normalizers[name] = simple_normalizer(name, BORDER_PROPS, check_inherit=False)

SHORTHAND_DEFAULTS = {
    'margin': '0',
    'padding': '0',
    'border-style': 'none',
    'border-width': '0',
    'border-color': 'currentColor',
    'border': 'none',
    'border-left': 'none',
    'border-right': 'none',
    'border-top': 'none',
    'border-bottom': 'none',
    'list-style': 'inherit',
    'font': 'inherit',
}

_safe_parser = None


def safe_parser():
    global _safe_parser
    if _safe_parser is None:
        import logging

        _safe_parser = CSSParser(loglevel=logging.CRITICAL, validate=False)
    return _safe_parser


def normalize_filter_css(props):
    ans = set()
    p = safe_parser()
    for prop in props:
        n = normalizers.get(prop, None)
        ans.add(prop)
        if n is not None and prop in SHORTHAND_DEFAULTS:
            dec = p.parseStyle(f'{prop}: {SHORTHAND_DEFAULTS[prop]}')
            cssvalue = dec.getPropertyCSSValue(dec.item(0))
            ans |= set(n(prop, cssvalue))
    return ans


def condense_edge(vals):
    edges = {x.name.rpartition('-')[-1]: x.value for x in vals}
    if len(edges) != 4 or set(edges) != {'left', 'top', 'right', 'bottom'}:
        return
    ce = {}
    for x, y in [('left', 'right'), ('top', 'bottom')]:
        if edges[x] == edges[y]:
            ce[x] = edges[x]
        else:
            ce[x], ce[y] = edges[x], edges[y]
    if len(ce) == 4:
        return ' '.join(ce[x] for x in ('top', 'right', 'bottom', 'left'))
    if len(ce) == 3:
        if 'right' in ce:
            return ' '.join(ce[x] for x in ('top', 'right', 'top', 'left'))
        return ' '.join(ce[x] for x in ('top', 'left', 'bottom'))
    if len(ce) == 2:
        if ce['top'] == ce['left']:
            return ce['top']
        return ' '.join(ce[x] for x in ('top', 'left'))


def simple_condenser(prefix, func):
    @wraps(func)
    def condense_simple(style, props):
        cp = func(props)
        if cp is not None:
            for prop in props:
                style.removeProperty(prop.name)
            style.setProperty(prefix, cp)

    return condense_simple


def condense_border(style, props):
    prop_map = {p.name: p for p in props}
    edge_vals = []
    for edge in EDGES:
        name = f'border-{edge}'
        vals = []
        for prop in BORDER_PROPS:
            x = prop_map.get(f'{name}-{prop}', None)
            if x is not None:
                vals.append(x)
        if len(vals) == 3:
            for prop in vals:
                style.removeProperty(prop.name)
            style.setProperty(name, ' '.join(x.value for x in vals))
            prop_map[name] = style.getProperty(name)
        x = prop_map.get(name, None)
        if x is not None:
            edge_vals.append(x)
    if len(edge_vals) == 4 and len({x.value for x in edge_vals}) == 1:
        for prop in edge_vals:
            style.removeProperty(prop.name)
        style.setProperty('border', edge_vals[0].value)


condensers = {
    'margin': simple_condenser('margin', condense_edge),
    'padding': simple_condenser('padding', condense_edge),
    'border': condense_border,
}


def condense_rule(style):
    expanded = {'margin-': [], 'padding-': [], 'border-': []}
    for prop in style.getProperties():
        for x, t in expanded.items():
            if prop.name and prop.name.startswith(x):
                t.append(prop)
                break
    for prefix, vals in expanded.items():
        if len(vals) > 1 and {x.priority for x in vals} == {''}:
            condensers[prefix[:-1]](style, vals)


def condense_sheet(sheet):
    for rule in sheet.cssRules:
        if rule.type == rule.STYLE_RULE:
            condense_rule(rule.style)


def test_normalization(return_tests=False):  # {{{
    import unittest
    from itertools import product

    from css_parser import parseStyle

    class TestNormalization(unittest.TestCase):
        longMessage = True
        maxDiff = None

        def test_font_normalization(self):
            def font_dict(expected):
                ans = {k: DEFAULTS[k] for k in font_composition} if expected else {}
                ans.update(expected)
                return ans

            for raw, expected in {
                'some_font': {'font-family': 'some_font'},
                'inherit': {k: 'inherit' for k in font_composition},
                '1.2pt/1.4 A_Font': {'font-family': 'A_Font', 'font-size': '1.2pt', 'line-height': '1.4'},
                'bad font': {'font-family': '"bad font"'},
                '10% serif': {'font-family': 'serif', 'font-size': '10%'},
                '12px "My Font", serif': {'font-family': '"My Font", serif', 'font-size': '12px'},
                'normal 0.6em/135% arial,sans-serif': {
                    'font-family': 'arial, sans-serif',
                    'font-size': '0.6em',
                    'line-height': '135%',
                    'font-style': 'normal',
                },
                'bold italic large serif': {
                    'font-family': 'serif',
                    'font-weight': 'bold',
                    'font-style': 'italic',
                    'font-size': 'large',
                },
                'bold italic small-caps larger/normal serif': {
                    'font-family': 'serif',
                    'font-weight': 'bold',
                    'font-style': 'italic',
                    'font-size': 'larger',
                    'line-height': 'normal',
                    'font-variant': 'small-caps',
                },
                '2em A B': {'font-family': '"A B"', 'font-size': '2em'},
            }.items():
                val = tuple(parseStyle(f'font: {raw}', validate=False))[0].propertyValue
                style = normalizers['font']('font', val)
                self.assertDictEqual(font_dict(expected), style, raw)

        def test_border_normalization(self):
            def border_edge_dict(expected, edge='right'):
                ans = {f'border-{edge}-{x}': DEFAULTS[f'border-{edge}-{x}'] for x in ('style', 'width', 'color')}
                for x, v in expected.items():
                    ans[f'border-{edge}-{x}'] = v
                return ans

            def border_dict(expected):
                ans = {}
                for edge in EDGES:
                    ans.update(border_edge_dict(expected, edge))
                return ans

            def border_val_dict(expected, val='color'):
                ans = {f'border-{edge}-{val}': DEFAULTS[f'border-{edge}-{val}'] for edge in EDGES}
                for edge in EDGES:
                    ans[f'border-{edge}-{val}'] = expected
                return ans

            for raw, expected in {
                'solid 1px red': {'color': 'red', 'width': '1px', 'style': 'solid'},
                '1px': {'width': '1px'},
                '#aaa': {'color': '#aaa'},
                '2em groove': {'width': '2em', 'style': 'groove'},
            }.items():
                for edge in EDGES:
                    br = f'border-{edge}'
                    val = tuple(parseStyle(f'{br}: {raw}', validate=False))[0].propertyValue
                    self.assertDictEqual(border_edge_dict(expected, edge), normalizers[br](br, val))

            for raw, expected in {
                'solid 1px red': {'color': 'red', 'width': '1px', 'style': 'solid'},
                '1px': {'width': '1px'},
                '#aaa': {'color': '#aaa'},
                'thin groove': {'width': 'thin', 'style': 'groove'},
            }.items():
                val = tuple(parseStyle('{}: {}'.format('border', raw), validate=False))[0].propertyValue
                self.assertDictEqual(border_dict(expected), normalizers['border']('border', val))

            for name, val in {
                'width': '10%',
                'color': 'rgb(0, 1, 1)',
                'style': 'double',
            }.items():
                cval = tuple(parseStyle(f'border-{name}: {val}', validate=False))[0].propertyValue
                self.assertDictEqual(border_val_dict(val, name), normalizers['border-' + name]('border-' + name, cval))

        def test_edge_normalization(self):
            def edge_dict(prefix, expected):
                return {f'{prefix}-{edge}': x for edge, x in zip(EDGES, expected)}

            for raw, expected in {
                '2px': ('2px', '2px', '2px', '2px'),
                '1em 2em': ('1em', '2em', '1em', '2em'),
                '1em 2em 3em': ('1em', '2em', '3em', '2em'),
                '1 2 3 4': ('1', '2', '3', '4'),
            }.items():
                for prefix in ('margin', 'padding'):
                    cval = tuple(parseStyle(f'{prefix}: {raw}', validate=False))[0].propertyValue
                    self.assertDictEqual(edge_dict(prefix, expected), normalizers[prefix](prefix, cval))

        def test_list_style_normalization(self):
            def ls_dict(expected):
                ans: dict = {f'list-style-{x}': DEFAULTS[f'list-style-{x}'] for x in ('type', 'image', 'position')}
                for k, v in expected.items():
                    ans[f'list-style-{k}'] = v
                return ans

            for raw, expected in {
                'url(http://www.example.com/images/list.png)': {'image': 'url(http://www.example.com/images/list.png)'},
                'inside square': {'position': 'inside', 'type': 'square'},
                'upper-roman url(img) outside': {'position': 'outside', 'type': 'upper-roman', 'image': 'url(img)'},
            }.items():
                cval = tuple(parseStyle(f'list-style: {raw}', validate=False))[0].propertyValue
                self.assertDictEqual(ls_dict(expected), normalizers['list-style']('list-style', cval))

        def test_filter_css_normalization(self):
            ae = self.assertEqual
            ae({'font'} | set(font_composition), normalize_filter_css({'font'}))
            for p in ('margin', 'padding'):
                ae({p} | {p + '-' + x for x in EDGES}, normalize_filter_css({p}))
            bvals = {f'border-{edge}-{x}' for edge in EDGES for x in BORDER_PROPS}
            ae(bvals | {'border'}, normalize_filter_css({'border'}))
            for x in BORDER_PROPS:
                sbvals = {f'border-{e}-{x}' for e in EDGES}
                ae(sbvals | {f'border-{x}'}, normalize_filter_css({f'border-{x}'}))
            for e in EDGES:
                sbvals = {f'border-{e}-{x}' for x in BORDER_PROPS}
                ae(sbvals | {f'border-{e}'}, normalize_filter_css({f'border-{e}'}))
            ae(
                {'list-style', 'list-style-image', 'list-style-type', 'list-style-position'},
                normalize_filter_css({'list-style'}),
            )

        def test_edge_condensation(self):
            for s, v in {
                (1, 1, 3): None,
                (1, 2, 3, 4): '2pt 3pt 4pt 1pt',
                (1, 2, 3, 2): '2pt 3pt 2pt 1pt',
                (1, 2, 1, 3): '2pt 1pt 3pt',
                (1, 2, 1, 2): '2pt 1pt',
                (1, 1, 1, 1): '1pt',
                ('2%', '2%', '2%', '2%'): '2%',
                tuple('0 0 0 0'.split()): '0',
            }.items():
                for prefix in ('margin', 'padding'):
                    css = {f'{prefix}-{x}': str(y) + 'pt' if isinstance(y, numbers.Number) else y for x, y in zip(('left', 'top', 'right', 'bottom'), s)}
                    css = '; '.join((f'{k}:{v}' for k, v in css.items()))
                    style = parseStyle(css)
                    condense_rule(style)
                    val = getattr(style.getProperty(prefix), 'value', None)
                    self.assertEqual(v, val)
                    if val is not None:
                        for edge in EDGES:
                            self.assertFalse(getattr(style.getProperty(f'{prefix}-{edge}'), 'value', None))

        def test_border_condensation(self):
            vals = 'red solid 5px'
            css = '; '.join(f'border-{edge}-{p}: {v}' for edge in EDGES for p, v in zip(BORDER_PROPS, vals.split()))
            style = parseStyle(css)
            condense_rule(style)
            for e, p in product(EDGES, BORDER_PROPS):
                self.assertFalse(style.getProperty(f'border-{e}-{p}'))
                self.assertFalse(style.getProperty(f'border-{e}'))
                self.assertFalse(style.getProperty(f'border-{p}'))
            self.assertEqual(style.getProperty('border').value, vals)
            css = '; '.join(f'border-{edge}-{p}: {v}' for edge in ('top',) for p, v in zip(BORDER_PROPS, vals.split()))
            style = parseStyle(css)
            condense_rule(style)
            self.assertEqual(css_text(style).rstrip(';'), f'border-top: {vals}')
            css += ';' + '; '.join(
                f'border-{edge}-{p}: {v}' for edge in ('right', 'left', 'bottom') for p, v in zip(BORDER_PROPS, vals.replace('red', 'green').split())
            )
            style = parseStyle(css)
            condense_rule(style)
            self.assertEqual(len(style.getProperties()), 4)
            self.assertEqual(style.getProperty('border-top').value, vals)
            self.assertEqual(style.getProperty('border-left').value, vals.replace('red', 'green'))

    tests = unittest.defaultTestLoader.loadTestsFromTestCase(TestNormalization)
    if return_tests:
        return tests
    unittest.TextTestRunner(verbosity=4).run(tests)


# }}}

if __name__ == '__main__':
    test_normalization()



#!/usr/bin/env python
# License: GPLv3 Copyright: 2012, Kovid Goyal <kovid at kovidgoyal.net>

from functools import wraps
from typing import Any

from calibre import prints
from calibre.constants import DEBUG
from calibre.devices.interface import DevicePlugin
from calibre.utils.localization import _


def debug(*args, **kwargs):
    if DEBUG:
        prints('MTP:', *args, **kwargs)


def synchronous(func):
    @wraps(func)
    def synchronizer(self, *args, **kwargs):
        with self.lock:
            return func(self, *args, **kwargs)

    return synchronizer


class MTPDeviceBase(DevicePlugin):
    name = 'MTP Device Interface'
    gui_name = _('MTP device')
    icon = 'devices/tablet.png'
    description = _('Communicate with MTP devices')
    author = 'Kovid Goyal'
    version = (1, 0, 0)
    prefs: dict[str, Any]

    def __init__(self, *args, **kwargs):
        DevicePlugin.__init__(self, *args, **kwargs)
        self.progress_reporter = None
        self.current_friendly_name = None
        self.report_progress = lambda x, y: None
        self.current_serial_num = None

    def reset(self, key='-1', log_packets=False, report_progress=None, detected_device=None):
        pass

    def set_progress_reporter(self, report_progress):
        self.report_progress = report_progress

    def get_gui_name(self):
        return getattr(self, 'current_friendly_name', self.gui_name)

    def is_usb_connected(self, devices_on_system, debug=False, only_presence=False):
        # We manage device presence ourselves, so this method should always
        # return False
        return False

    @property
    def save_template(self) -> str:
        raise NotImplementedError

    def filesystem_callback(self, msg: str) -> None:
        pass

    def is_folder_ignored(self, storage_or_storage_id, path) -> bool:
        return False

    def build_template_regexp(self):
        from calibre.devices.utils import build_template_regexp

        return build_template_regexp(self.save_template)

    def is_customizable(self):
        return True



# License: GPLv3 Copyright: 2009, John Schember <john@nachtimwald.com>

from calibre.ebooks.conversion.config import OPTIONS
from calibre.gui2.convert import Widget
from calibre.gui2.convert.txt_output_ui import Ui_Form
from calibre.utils.localization import _


class PluginWidget(Widget, Ui_Form):
    TITLE = _('TXT output')
    HELP = _('Options specific to') + ' TXT ' + _('output')
    COMMIT_NAME = 'txt_output'
    ICON = 'mimetypes/txt.png'

    def __init__(self, parent, get_option, get_help, db=None, book_id=None):
        Widget.__init__(self, parent, OPTIONS['output']['txt'])
        self.db, self.book_id = db, book_id
        for x in get_option('newline').option.choices:
            self.opt_newline.addItem(x)
        for x in get_option('txt_output_formatting').option.choices:
            self.opt_txt_output_formatting.addItem(x)
        self.initialize_options(get_option, get_help, db, book_id)
        if self.COMMIT_NAME != 'txt_output':
            self.image_note_label.setText(
                _(
                    'Note that for images to be preserved, the formatting option above must be markdown or textile'
                    ' and also enable the setting to not remove image references.'
                )
            )

    def break_cycles(self):
        Widget.break_cycles(self)
